/*******************************************************************************
filter.cpp

example of a filtered subscription of FT_C_PRICE_EXEC class

*******************************************************************************/

#include "ftsome.h"

FTRevision appVer = {1,0,0};
unsigned long applSign = 732;

FTContext ft_ctx = NULL;
FTAPI_CONF_REC conf = {0};

//METAMARKET CONNECTION DESCRIPTOR
FTConversation_DESC_REC mm_conv_desc;

//////////////////////////////// FILTER ////////////////////////////////////////

static int filter_created = 0;
static int filter_set = 0;
static long filter_id = 0;

int eval_filterCreated(void *ud){
    return filter_created;
}

int eval_filterSet(void *ud){
    return filter_set;
}

void clbk_mm_FTFilterCreateProc(long ConversationID, unsigned long ReqID, long FilterID, FTFilterCreateEnum Res){
    FTTraceAppl(FTLevel_Normal, "filter created [%s]", Res == FTFilterCreateOK ? "OK" : "KO");
    filter_id = FilterID;
    filter_created = 1;
}

void clbk_mm_FTFilterSetProc(long ConversationID, unsigned long ReqID, long FilterID, FTFilterSetEnum Res){
    FTTraceAppl(FTLevel_Normal, "filter set [%s]", Res == FTFilterSetOK ? "OK" : "KO");
    filter_set = 1;
}

////////////////////////////////APPLICATION LOGIC ON SUBSCRIBED CLASSES ////////////////////////////////////////

long handleFT_C_PRICE_EXEC(long ConversationID, unsigned long ReqID, FTEventTypeEnum EventType, FTNotifySubscribeResStruct *ResStruct){
    FT_C_PRICE_EXECPtr obj = (FT_C_PRICE_EXECPtr)ResStruct->MarketObject;
    switch (ResStruct->EntityAction) {
        case FTEntityKIL:
            //set new class version
            if (ResStruct->KeyID <= 0 && !ResStruct->MarketObject){
                // This means a class reset at the server level.
                // Every entity received previously have to be invalidated!
                FTTraceAppl(FTLevel_Normal, "FTEntityKIL Action - Class Reset Signal - New version: %d", (ResStruct->TimeStamp)[0]);
            }
            else{
                FTTraceAppl(FTLevel_Normal, "FTEntityKIL Action - KeyID [%d]", ResStruct->KeyID);
            }
            break;
        case FTEntityDEL:
            FTTraceAppl(FTLevel_Normal, "EntityDEL Action - KeyID [%d]", ResStruct->KeyID);
            break;
        case FTEntityADD:
        case FTEntityRWT:
            FTTraceAppl(FTLevel_Normal, "EntityADD/RWT Action - KeyID [%d]", ResStruct->KeyID);
            FTTraceAppl(FTLevel_Normal, "FT_C_PRICE_EXEC [BID QTY:%f BID PR:%f]", obj->Qty, obj->Price);
            break;
    }
    return 0;
}



/////////////////////////////////SUBSCRIPTION CALLBACKS////////////////////////////////////////

static void NotifyClass(long ConversationID, unsigned long ReqID, FTEventTypeEnum EventType, FTNotifySubscribeResStruct *ResStruct){
    switch(EventType){
    case FTConnectionBroken:
        FTTraceAppl(FTLevel_Error, "NotifyClass: connection broken (%d,%d)" , ConversationID, ReqID);
        break;
    case FTClassIdle:
        FTTraceAppl(FTLevel_Normal, "idle received on class %d", ResStruct->ClassID);
        break;
    case FTClassInstance: {
            switch(ResStruct->ClassID){
                case FT_C_PRICE_EXEC_ID:
                    handleFT_C_PRICE_EXEC(ConversationID, ReqID, EventType, ResStruct);
                    break;
                default:
                    break;
            }
        }
    default:
        break;
    }
}

static void SubscribeResProc(long convID, unsigned long ReqID, long ResultCode, FTSubscribeResStruct *resStruct){
    FTTraceAppl(FTLevel_Normal, "-------- FTOpenSubscribeResProc: classID [%lu], SubscriptionID [%lu]", resStruct->ClassID, resStruct->SubscriptionID);
    FTTraceAppl(FTLevel_Normal, "-------- Result: %s", FTGetErrorString(ResultCode));

    if (ResultCode == FTOK){
        FTTraceAppl(FTLevel_Normal, "-------- Status: %s", (resStruct->SubscribeStatus==FTSubscribeStart)? "Start" : "Stop");
        FTTraceAppl(FTLevel_Normal, "-------- New Table version: %lu; To-Reset signal: %d", resStruct->ClassVersion, resStruct->ResetClass);

        if (resStruct->ResetClass){
            // In this case your request can't be served in incremental manner!
            // Any previous record received have to be considered invalidated and
            // the server will resend all the class content.
            FTTraceAppl(FTLevel_Normal, "-------- Class Reset - Class %d: New version [%lu]", resStruct->ClassID, resStruct->ClassVersion);
        }
    }
}

int price_exec_subscribed = 0;

int subscribePriceExec(){
    long res = 0;
    char filter[2048] = {'\0'};
    int idx = 0;

	/******************************************************************************************
     *	Filter.set value syntax
     *
     *  Add a filter key:
     *  A!KeyID:NumSegs!KeySeg1;KeySeg2;KeySeg3\tA!KeyID:NumSegs!KeySeg1;KeySeg2;KeySeg3\0
     *
     *  Del a filter key:
     *  D!KeyID:NumSegs!KeySeg1;KeySeg2;KeySeg3\tD!KeyID:NumSegs!KeySeg1;KeySeg2;KeySeg3\0
     *
     *	Reset all filter keys:
     *	R\0
     *
     *  Examples:
     *
     *	R\tA!1!56700;Pippo\0
     *	D!1!56700;Pippo\tA!1!45700;Pluto\0
     *
     ******************************************************************************************/

    // Multiple sets are evaluated in 'or' manner

    char ftsecid[256];
    FILE *fp = fopen("filter", "r");
	if(fp){
    while(!feof(fp)){
	    if(fgets(ftsecid, 256, fp)){
		    if(IsBlankLine(ftsecid) || IsCommentLine(ftsecid)){
			    continue;
		    }else{
                ftsecid[strlen(ftsecid)-1] = '\0';
                idx += sprintf(&filter[idx], "A!1:1!%s\t", ftsecid);
		    }
	    }
    }
	fclose(fp);
	}else{
        FTTraceAppl(FTLevel_Error, "cannot open filter file.");
		return -1;
	}

    res = FTFilterCreate(mm_conv_desc.ConversationID,
                         244,
                         FT_C_PRICE_EXEC_ID,
                         GIDPFILTER_TYPE,
                         0, 0,
                         clbk_mm_FTFilterCreateProc);
    
    if(res){
        return -1;
    }

    FTRunCycleUntilCondition(eval_filterCreated, NULL);

    res = FTFilterSet(mm_conv_desc.ConversationID,
                        filter_id, 
                        476,
                        strlen(filter),
                        filter,
                        clbk_mm_FTFilterSetProc);

    if(res){
        return -1;
    }

    FTRunCycleUntilCondition(eval_filterSet, NULL);

    FTTraceAppl(FTLevel_Normal, "subscribing FT_C_PRICE_EXEC class..");

    res = FTStartSubscribeClassExt(mm_conv_desc.ConversationID,
                                        FTAckUnrequired,
                                        FTFlowLast,
                                        FT_C_PRICE_EXEC_ID,
                                        0, 0,
                                        FTQueryAll,
                                        FT_C_PRICE_EXECKey,
                                        0, 0,
                                        SubscribeResProc, 
                                        NotifyClass,
                                        865,
                                        filter_id,
                                        0);
    
    if(res <= 0){
        return -1;
    }

    price_exec_subscribed = 1;
    return 0;
}


/////////////////////////////////MM OPEN-CONV CALLBACKS////////////////////////////////////////
void MMOpenConvCallback(long ConversationID, long ResultCode, FTOpenResStruct *ResStruct){
    long res = 0;
    if((mm_conv_desc.open_conv_ResultCode = ResultCode) == FTOK){
        FTTraceAppl(FTLevel_Normal, "MMOpenConvCallback: MetaMarket connection is ON!");
        FTTraceAppl(FTLevel_Normal, "MMOpenConvCallback: Market Revision: V:%d.%d.%d",   
            ResStruct->MarketRevision[0],
            ResStruct->MarketRevision[1],
            ResStruct->MarketRevision[2]);

        mm_conv_desc.ConversationID = ConversationID;

        subscribePriceExec();

   }else{
        FTTraceAppl(FTLevel_Normal, "metamarket connection FAILED! (%d)", ResultCode);
        if(ResultCode == FTInvalidCID){
            FTTraceAppl(FTLevel_Normal, "CID was already used on platform we try to connect with another one..");
                if((res = ConnectToServer(&conf, ft_ctx, 999)) < 0){
                FTTraceAppl(FTLevel_Normal, "ConnectToServer failed [result: %d] program exiting..", res);
                //error
                FTShutdown();
                exit(-1);
            }
        }
    }
}

void MMBrokenConvCallback(long ConversationID, FTBrokenTypeEnum BrokenType){
    FTTraceAppl(FTLevel_Error, "MMBrokenConvCallback: metamarket connection broken (%d, %d)" , ConversationID, BrokenType);
}


/////////////////////////////////MAIN////////////////////////////////////////
///
int main(int argc, char *argv[]){
    long result = FTOK;

    if(LoadConfigByFile(&conf)){
        printf("PARAMETERS ERROR! exiting..\n");
        return 1;
    }

    //create an ft-context and set its properties
    ft_ctx = FTCreateContext();
    FTSetContextAttribute(ft_ctx, CTX_COMPRESSED, "ZIP");
    FTSetContextAttribute(ft_ctx, CTX_APPL_AUTHFILE, conf.license_file);

    //initialize ft-framework
    if(InitializeFTAPI(&conf, ft_ctx) != FTOK){
        //error
        FTShutdown();
        return 1;
    }

    conf.OpenResProc = MMOpenConvCallback;
    conf.BrokenProc = MMBrokenConvCallback;
    conf.UserType = FTUserTrader;

    //remember that connection will not be done until you call at least one FTRun() call.
    if(ConnectToServer(&conf, ft_ctx, 999) < 0){
        //error
        FTShutdown();
        return 1;
    }

/***********************************************************************************
** MAIN EVENT LOOP **
 Calling FTRun() ensures that all requests (subscriptions, transactions, queries ..) 
 will be handled by FTAPI durig program lifecycle.
***********************************************************************************/
    unsigned long times = 0; 
    do{
        if(FTRun() != FTOK){
            FTShutdown();
            return 1;
        }
        times++;
        if(!(times % 10)){
            FTTraceAppl(FTLevel_Full, "FTRun called %lu times", times);
        }
        FTSocketSelect();
    }while(1);

    //unreachable code
    return 0;
}

